#!/usr/bin/env python3

from utils.kernels_derivatives import Gaussian
import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted


# VKOGA implementation
class ecVKOGA(BaseEstimator):

    def __init__(self, kernel=Gaussian(), kernel_par=1,
                 verbose=True, n_report=10,
                 greedy_type='p_greedy', reg_par=0, restr_par=0,
                 tol_f=1e-10, tol_p=1e-10, max_iter=10):

        # Set the verbosity on/off
        self.verbose = verbose

        # Set the frequency of report
        self.n_report = n_report

        # Set the params defining the method
        self.kernel = kernel
        self.kernel_par = kernel_par
        self.greedy_type = greedy_type
        self.reg_par = reg_par

        if reg_par > 0:
            print('reg_para > 0 is not yet implemented correctly!')
        self.restr_par = restr_par

        # Set the stopping values
        self.max_iter = max_iter
        self.tol_f = tol_f
        self.tol_p = tol_p

        # Set the output dimension
        self.q = None

    def selection_rule(self, f, p):
        if self.restr_par > 0:
            p_ = np.max(p)
            restr_idx = np.nonzero(p >= self.restr_par * p_)[0]
        else:
            restr_idx = np.arange(len(p))

        f = np.sum(f ** 2, axis=1)
        if self.greedy_type == 'f_greedy':
            idx = np.argmax(f[restr_idx])
            idx = restr_idx[idx]
            f_max = np.max(f)
            p_max = np.max(p)
        elif self.greedy_type == 'fp_greedy':  # I'm not sure whether the optimality of fp-greedy still holds
            idx = np.argmax(f[restr_idx] / p[restr_idx])
            idx = restr_idx[idx]
            f_max = np.max(f)
            p_max = np.max(p)
        elif self.greedy_type == 'p_greedy':
            f_max = np.max(f)
            idx = np.argmax(p)
            idx = restr_idx[idx]
            p_max = p[idx]

        return idx, f_max, p_max

    def fit(self, X, y, weight=None, flag_early_stopping=False, flag_select_origin_first=True):
        # weight (of shape 1 x y.shape[1]) allows to weight the dimensions of the target vector
        if weight is None:
            weight = 1

        # flag needed to break out of loop if negative power values occur
        flag_break = False

        # Check the dataset
        X, y = check_X_y(X, y, multi_output=True)

        # Initialize the convergence history (cold start)
        self.train_hist = {}
        self.train_hist['n'] = []
        self.train_hist['f'] = []
        self.train_hist['p'] = []

        # Initialize the residual (i.e. the derivatives of the residual of the potential)
        y = np.array(y)
        if len(y.shape) == 1:
            y = y[:, None]

        # Get the data dimension
        N, q = y.shape
        self.q = q

        self.max_iter = min(self.max_iter, N)

        self.kernel.set_params(self.kernel_par)

        # Check compatibility of restriction
        if self.greedy_type == 'p_greedy':
            self.restr_par = 0
        if not self.reg_par == 0:
            self.restr_par = 0

        self.indI_ = []
        notIndI = list(range(N))

        # Initialize arrays to store Newton basis and coefficients
        Vx = np.zeros((N, self.max_iter * q))
        c = np.zeros(self.max_iter * q)

        # Initialize a 3D array which stores the derivatives of the Newton basis
        der_Vx = np.zeros((q, N, self.max_iter * q))

        # Initialize an array to store squared power function values in each direction
        # p = np.matlib.repmat(self.kernel.dd_diagonal(X)[:, None] + self.reg_par, 1, q)
        p = self.kernel.dd_diagonal(X)

        # Initialize Cut matrix (transpose of C matrix, such that it is lower triangular
        self.Cut_ = np.zeros((self.max_iter * q, self.max_iter * q))

        self.print_message('begin')
        # Iterative selection of new points
        for n_point in range(self.max_iter):
            # prepare
            self.train_hist['n'].append(n_point + 1)
            self.train_hist['f'].append([])
            self.train_hist['p'].append([])

            # Apply selection rule (we take the l2 norm for the power values)
            if n_point == 0 and flag_select_origin_first:  # select the origin at the beginning!
                idx = 0
                self.train_hist['f'][n_point] = np.max(np.sum((y / weight) ** 2, axis=1))
                self.train_hist['p'][n_point] = np.max(np.linalg.norm(p[notIndI, :], 2, 1)[:, None])
            else:  # standard selection for later points
                idx, self.train_hist['f'][n_point], self.train_hist['p'][n_point] = \
                    self.selection_rule(y / weight, np.linalg.norm(p[notIndI, :], 2, 1)[:, None])

            # add the current index
            self.indI_.append(notIndI[idx])

            # check if the tolerances are reacheded
            if self.train_hist['f'][n_point] <= self.tol_f:
                n_point = n_point - 1
                self.print_message('end')
                break
            if self.train_hist['p'][n_point] <= self.tol_p:
                n_point = n_point - 1
                self.print_message('end')
                break

            # For each point use all derivative
            for idx_dim in range(q):
                # Compute the next Newton basis element
                Vx[notIndI, n_point * q + idx_dim] = \
                    (self.kernel.deval(X[notIndI, :], X[[self.indI_[n_point]], :], idx_dim)
                     - Vx[notIndI, 0:n_point * q + idx_dim + 1] @
                     der_Vx[idx_dim, [self.indI_[n_point]], 0:n_point * q + idx_dim + 1].transpose()).reshape(-1)

                # Normalize the Newton basis element (if negative values occur: break)
                if p[self.indI_[n_point], idx_dim] < 0:
                    # n_point = n_point - 5
                    self.print_message('end')
                    flag_break = True

                    break

                Vx[notIndI, n_point * q + idx_dim] = \
                    Vx[notIndI, n_point * q + idx_dim] / np.sqrt(p[self.indI_[n_point], idx_dim])

                # Calculate all q derivatives of the previously computed Newton basis element
                for idx_dim2 in range(q):
                    der_Vx[idx_dim2, notIndI, n_point * q + idx_dim] = \
                        (self.kernel.ddeval(X[notIndI, :], X[[self.indI_[n_point]], :], idx_dim2, idx_dim)
                         - der_Vx[idx_dim2, notIndI, 0:n_point * q + idx_dim + 1] @
                         der_Vx[idx_dim, [self.indI_[n_point]], 0:n_point * q + idx_dim + 1].transpose()).reshape(-1)

                    der_Vx[idx_dim2, notIndI, n_point * q + idx_dim] = \
                        der_Vx[idx_dim2, notIndI, n_point * q + idx_dim] / \
                        np.sqrt(p[self.indI_[n_point], idx_dim])

                # Calculate next row of Cut and then update it
                Cut_new_row = np.ones(n_point * q + idx_dim + 1)  # like this the last entry is set to one
                Cut_new_row[0:n_point * q + idx_dim] = -der_Vx[idx_dim, self.indI_[n_point], 0:n_point * q + idx_dim] \
                                                       @ self.Cut_[:n_point * q + idx_dim:, :n_point * q + idx_dim]

                self.Cut_[n_point * q + idx_dim, :n_point * q + idx_dim + 1] = \
                    Cut_new_row / der_Vx[idx_dim, self.indI_[n_point], n_point * q + idx_dim]

                # compute the (n_point*q+idx_dim)-th coefficient
                c[n_point * q + idx_dim] = \
                    y[self.indI_[n_point], idx_dim] / np.sqrt(p[self.indI_[n_point], idx_dim])

                # update all q Power functions
                p[notIndI, :] = p[notIndI, :] - der_Vx[:, notIndI, n_point * q + idx_dim].transpose() ** 2

                # update the residuals (i.e. the derivatives of the residual of the potential)
                y[notIndI, :] = y[notIndI, :] - \
                                der_Vx[:, notIndI, n_point * q + idx_dim].transpose() * c[n_point * q + idx_dim]

            if flag_break:
                print('Break due do negative power values.')
                break

            # remove the nth index from the dictionary
            notIndI.pop(idx)

            # Report some data every now and then
            if n_point % self.n_report == 0:
                self.print_message('track')

        else:
            self.print_message('end')

        # Early stopping:
        if flag_early_stopping:
            idx_argmin = np.argmin(np.asarray(self.train_hist['f']))
            print('flag_early_stopping removes {} centers.'.format(
                n_point + 1 - idx_argmin))
            n_point = idx_argmin - 1  # next lines use n_point + 1

        # Define coefficients and centers
        c = c[:(n_point + 1) * q]  # recall we used n_point = n_point - 1 above
        self.Cut_ = self.Cut_[0:(n_point + 1) * q, 0:(n_point + 1) * q]
        self.indI_ = self.indI_[0:(n_point + 1)]
        self.coef_ = self.Cut_.transpose() @ c
        self.ctrs_ = X[self.indI_, :]

        print('{} / {}'.format(n_point, N))

        return self

    def predict(self, X):
        # This one does NOT predict the potential, but the forces (i.e. the derivatives derived from the potential)

        # Check if fit has been called
        check_is_fitted(self, 'coef_')

        # Validate the input
        X = check_array(X)

        # Evaluate the model (for this loop over the output dimensions) (ToDo: improve this part of code .. :/)
        coeffs = self.coef_.reshape(-1, self.q)

        list_to_concatenate = []
        # Loop over dimensions to calculate all outputs
        for idx_dim in range(self.q):
            array_sum = np.zeros((X.shape[0], 1))

            # For each center point there are q Newton Riesz representer, thus loop
            for idx_dim2 in range(self.q):
                array_to_add = self.kernel.ddeval(X, self.ctrs_, idx_dim, idx_dim2) @ coeffs[:, idx_dim2]

                array_sum += array_to_add.reshape(-1, 1)

            list_to_concatenate.append(array_sum)

        # finally horizontal-stack the single dimensions
        return np.hstack(list_to_concatenate)

    def predict_tmp(self, X):
        # This one does NOT predict the potential, but the forces (i.e. the derivatives derived from the potential)

        # Check if fit has been called
        check_is_fitted(self, 'coef_')

        # Validate the input
        X = check_array(X)

        # Calculate all outputs
        result = self.kernel.ddeval_tmp(X, self.ctrs_) @ self.coef_.reshape(-1, self.q).flatten(order='F')

        return result.reshape(self.q, -1).transpose()

    def print_message(self, when):

        if self.verbose and when == 'begin':
            print('')
            print('*' * 30 + ' [VKOGA] ' + '*' * 30)
            print('Training model with')
            print('       |_ kernel              : %s' % self.kernel)
            print('       |_ regularization par. : %2.2e' % self.reg_par)
            print('       |_ restriction par.    : %2.2e' % self.restr_par)
            print('')

        if self.verbose and when == 'end':
            print('Training completed with')
            print('       |_ selected points     : %8d / %8d' % (self.train_hist['n'][-1], self.max_iter))
            print('       |_ train residual      : %2.2e / %2.2e' % (self.train_hist['f'][-1], self.tol_f))
            print('       |_ train power fun     : %2.2e / %2.2e' % (self.train_hist['p'][-1], self.tol_p))

        if self.verbose and when == 'track':
            print('Training ongoing with')
            print('       |_ selected points     : %8d / %8d' % (self.train_hist['n'][-1], self.max_iter))
            print('       |_ train residual      : %2.2e / %2.2e' % (self.train_hist['f'][-1], self.tol_f))
            print('       |_ train power fun     : %2.2e / %2.2e' % (self.train_hist['p'][-1], self.tol_p))


# %% Utilities to
import pickle


def save_object(obj, filename):
    with open(filename, 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)


def load_object(filename):
    with open(filename, 'rb') as input:
        obj = pickle.load(input)
    return obj
